/*
 * Simulator of microcontrollers (avr.cc)
 *
 * Copyright (C) 1999,99 Drotos Daniel, Talker Bt.
 * 
 * To contact author send email to drdani@mazsola.iit.uni-miskolc.hu
 *
 */

/* This file is part of microcontroller simulator: ucsim.

UCSIM is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

UCSIM is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with UCSIM; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA. */
/*@1@*/

#include "ddconfig.h"

#include <stdio.h>

// prj
#include "pobjcl.h"

// sim
#include "simcl.h"

// local
#include "portcl.h"
#include "avrcl.h"
#include "glob.h"


/*
 * Base type of AVR microcontrollers
 */

cl_avr::cl_avr(class cl_sim *asim):
  cl_uc(asim)
{
  type= CPU_AVR;
}

int
cl_avr::init(void)
{
  cl_uc::init(); /* Memories now exist */
  
}

char *
cl_avr::id_string(void)
{
  return("unspecified AVR");
}


/*
 * Making elements of the controller
 */

long
cl_avr::get_mem_size(enum mem_class type)
{
  if (type == MEM_SFR)
    return(0);
  return(cl_uc::get_mem_size(type));
}

int
cl_avr::get_mem_width(enum mem_class type)
{
  if (type == MEM_ROM)
    return(16);
  return(cl_uc::get_mem_width(type));
}

void
cl_avr::mk_hw_elements(void)
{
  class cl_base *o;
  /* t_uc::mk_hw() does nothing */
  hws->add(o= new cl_port(this));
  o->init();
}


/*
 * Help command interpreter
 */

struct dis_entry *
cl_avr::dis_tbl(void)
{
  //FIXME
  return(0);
}

struct name_entry *
cl_avr::sfr_tbl(void)
{
  return(sfr_tabl);
}

struct name_entry *
cl_avr::bit_tbl(void)
{
  //FIXME
  return(0);
}

/*
 * Execution
 */

int
cl_avr::exec_inst(bool use_breaks)
{
  TYPE_UWORD code;

  code= fetch();
  tick(1);
  switch (code)
    {
    case 0x9409:
      return(ijmp());
    case 0x9419:
      return(eijmp());
    case 0x9509:
      return(icall());
    case 0x9519:
      return(eicall());
    case 0x9508:
      return(ret());
    case 0x9518:
      return(reti());
    case 0x95c8:
      return(lmp());
    case 0x95d8:
      return(elmp());
    case 0x95e8:
      return(smp());
    case 0x95f8:
      return(esmp());
    case 0x9408:
      return(sec());
    case 0x9488:
      return(clc());
    case 0x9428:
      return(sen());
    case 0x94a8:
      return(cln());
    case 0x9418:
      return(sez());
    case 0x9498:
      return(clz());
    case 0x9478:
      return(sei());
    case 0x94f8:
      return(cli());
    case 0x9448:
      return(ses());
    case 0x94c8:
      return(cls());
    case 0x9438:
      return(sev());
    case 0x94b8:
      return(clv());
    case 0x9468:
      return(set());
    case 0x94e8:
      return(clt());
    case 0x9458:
      return(seh());
    case 0x94d8:
      return(clh());
    case 0x0000:
      return(nop());
    case 0x9588:
      return(sleep());
    case 0x95a8:
      return(wdr());
    }
  switch (code & 0xf000)
    {
    case 0x3000: return(cpi_Rd_K());
    case 0x4000: return(sbci_Rd_K());
    case 0x5000: return(subi_Rd_K());
    case 0x6000: return(ori_Rd_K());
    case 0x7000: return(andi_Rd_K());
    case 0xc000: return(rjmp_k());
    case 0xd000: return(rcall_k());
    case 0xe000: return(ldi_Rd_K());
    }
  switch (code & 0xf000)
    {
    case 0x0000:
      // 0x0...
      switch (code & 0xfc00)
	{
	case 0x0000:
	  switch (code & 0xff00)
	    {
	    case 0x0100: return(movw_Rd_Rr());
	    case 0x0200: return(muls_Rd_Rr());
	    case 0x0300:
	      switch (code & 0xff88)
		{
		case 0x0300: return(mulsu_Rd_Rr());
		case 0x0308: return(fmul_Rd_Rr());
		case 0x0380: return(fmuls_Rd_Rr());
		case 0x0388: return(fmulsu_Rd_Rr());
		}
	    }
	case 0x0400: return(cpc_Rd_Rr());
	case 0x0800: return(sbc_Rd_Rr());
	case 0x0c00: return(add_Rd_Rr());
	}
    case 0x1000:
      // 0x1...
      switch (code & 0xfc0)
	{
	case 0x1000: return(cpse_Rd_Rr());
	case 0x1400: return(cp_Rd_Rr());
	case 0x1800: return(sub_Rd_Rr());
	case 0x1c00: return(adc_Rd_Rr());
	}
    case 0x2000:
      // 0x2...
      switch (code & 0xfc00)
	{
	case 0x2000: return(and_Rd_Rr());
	case 0x2400: return(eor_Rd_Rr());
	case 0x2800: return(or_Rd_Rr());
	case 0x2c00: return(mov_Rd_Rr());
	}
    case 0x8000:
      // 0x8...
      switch (code &0xf208)
	{
	case 0x8000: return(ldd_Rd_Z_q());
	case 0x8008: return(ldd_Rd_Y_q());
	case 0x8200: return(st_Z_q_Rr());
	case 0x8208: return(std_Y_q_Rr());
	}
    case 0x9000:
      // 0x9...
      switch (code & 0xfc00)
	{
	case 0x9000:
	  switch (code & 0xfe0f)
	    {
	    case 0x9000: return(lds_Rd_k());
	    case 0x9001: return(ld_Rd_Z$());
	    case 0x9002: return(ld_Rd_$Z());
	    case 0x9004: return(lpm_Rd_Z());
	    case 0x9005: return(lpm_Rd_Z$());
	    case 0x9006: return(elpm_Rd_Z());
	    case 0x9007: return(elpm_Rd_Z$());
	    case 0x9009: return(ld_Rd_Y$());
	    case 0x900a: return(ld_Rd_$Y());
	    case 0x900c: return(ld_Rd_X());
	    case 0x900d: return(ld_Rd_X$());
	    case 0x900e: return(ld_Rd_$X());
	    case 0x900f: return(pop_Rd());
	    case 0x9200: return(sts_k_Rr());
	    case 0x9201: return(st_Z$_Rr());
	    case 0x9202: return(st_$Z_Rr());
	    case 0x9209: return(st_Y$_Rr());
	    case 0x920a: return(st$Y_Rr());
	    case 0x920c: return(st_X_Rr());
	    case 0x920d: return(st_X$_Rr());
	    case 0x920e: return(st_$X_Rr());
	    case 0x920f: return(push_Rr());
	    }
	case 0x9400:
	  switch (code & 0xfe0f)
	    {
	    case 0x9400: return(com_Rd());
	    case 0x9401: return(neg_Rd());
	    case 0x9402: return(swap_Rd());
	    case 0x9403: return(inc_Rd());
	    case 0x9405: return(asr_Rd());
	    case 0x9406: return(lsr_Rd());
	    case 0x9407: return(ror_Rd());
	    case 0x940a: return(dec_Rd());
	    case 0x940c: case 0x940d: return(jmp_k());
	    case 0x940e: case 0x940f: return(call_k());
	    }
	case 0x9800:
	  switch (code & 0xff00)
	    {
	    case 0x9800: return(cbi_A_b());
	    case 0x9900: return(sbic_A_b());
	    case 0x9a00: return(sbi_A_b());
	    case 0x9b00: return(sbis_A_b());
	    }
	case 0x9c00: return(mul_Rd_Rr());
	}
    case 0xa000:
      // 0xa...
      switch (code &0xf208)
	{
	case 0xa000: return(ldd_Rd_Z_q());
	case 0xa008: return(ldd_Rd_Y_q());
	case 0xa200: return(st_Z_q_Rr());
	case 0xa208: return(std_Y_q_Rr());
	}
    case 0xb000:
      // 0xb...
      switch (code & 0xf800)
	{
	case 0xb000: return(in_Rd_A());
	case 0xb800: return(out_A_Rr());
	}
    case 0xf000:
      // 0xf...
      switch (code & 0xfc00)
	{
	case 0xf000: return(brbs_s_k());
	case 0xf400: return(brbc_s_k());
	case 0xf800: case 0xfc00:
	  switch (code & 0xfe08)
	    {
	    case 0xf800: return(bld_Rd_b());
	    case 0xfa00: return(bst_Rd_b());
	    case 0xfc00: return(sbrc_Rr_b());
	    case 0xfe00: return(sbrs_Rr_b());
	    }
	}
    }
  if (PC)
    PC--;
  else
    PC= get_mem_size(MEM_ROM)-1;
  ticks-= clock_per_cycle();
  sim->stop(resINV_INST);
  return(0);
}


/* End of avr.cc */
